
local scalefactor = 20


local INTERACTABLE = {
	[MT_STARPOST] = true,
	[MT_SIGN] = true,
}

local IGNORE = {
	[MT_CYBRAKDEMON_TARGET_DOT] = true,
	[MT_LASER] = true,
	[MT_EGGMOBILE_FIRE] = true,
	[MT_EMBLEM] = true,
	[MT_EGGMOBILE_TARGET] = true,
	[MT_EGGMOBILE_FIRE] = true,
	[MT_JETFUME1] = true,
	[MT_MINECART] = true,
	[MT_MINECARTSEG] = true,
	[MT_MINECARTSPAWNER] = true,
	[MT_MINECARTEND] = true,
	[MT_MINECARTENDSOLID] = true,
	[MT_MINECARTSIDEMARK] = true,
	[MT_MINECARTSPARK] = true,
	[MT_MINECARTSWITCHPOINT] = true,
	[MT_EXTRALARGEBUBBLE] = true,
	[MT_EGGSHIELD] = true,
	[MT_SMALLMACE] = true,
	[MT_BIGMACE] = true
}

local ALWAYS = {
	[MT_RING] = true,
	[MT_FLINGRING] = true,
	[MT_TOKEN] = true
}

-- objects that will be checked for proximity to the player
-- because they don't normally have collison checks
-- they will also be treated as having a different size than
-- the have defined because lots of these objects have way
-- larger radius or height than they logically should have
-- for how large the sprite is
local SCENERY = {
	-- checkpoint
	[MT_STARPOST] = 20,
	-- springs
	[MT_YELLOWSPRING] = 50,
	[MT_REDSPRING] = 50,
	[MT_BLUESPRING] = 50,
	[MT_YELLOWDIAG] = 50,
	[MT_REDDIAG] = 50,
	[MT_BLUEDIAG] = 50,
	[MT_YELLOWHORIZ] = 50,
	[MT_REDHORIZ] = 50,
	[MT_BLUEHORIZ] = 50,
	-- flickies
	[MT_FLICKY_01] = 1,
	[MT_FLICKY_02] = 1,
	[MT_FLICKY_03] = 1,
	[MT_FLICKY_04] = 1,
	[MT_FLICKY_05] = 1,
	[MT_FLICKY_06] = 1,
	[MT_FLICKY_07] = 1,
	[MT_FLICKY_08] = 1,
	[MT_FLICKY_09] = 1,
	[MT_FLICKY_10] = 1,
	[MT_FLICKY_11] = 1,
	[MT_FLICKY_12] = 1,
	[MT_FLICKY_13] = 1,
	[MT_FLICKY_14] = 1,
	[MT_FLICKY_15] = 1,
	[MT_FLICKY_16] = 1,
	[MT_SECRETFLICKY_01] = 1,
	[MT_SECRETFLICKY_02] = 1,
	-- Tutorial Zone scenery
	[MT_TUTORIALPLANT] = 26,
	[MT_TUTORIALLEAF] = 26,
	[MT_TUTORIALFLOWER] = 26,
	[MT_TUTORIALFLOWERF] = 26,
	-- Green Flower Zone scenery
	[MT_GFZFLOWER1] = 6,
	[MT_GFZFLOWER2] = 6,
	[MT_GFZFLOWER3] = 6,
	-- Techno Hill Zone scenery
	[MT_THZFLOWER1] = 6,
	[MT_THZFLOWER2] = 6,
	[MT_THZFLOWER3] = 6,
	[MT_THZTREE] = 6,
	[MT_THZTREEBRANCH] = 6,
	[MT_ALARM] = 6,
	-- Deep Sea Zone scenery
	[MT_SEAWEED] = 6,
	[MT_KELP] = 32,
	[MT_ANIMALGAETOP] = 32,
	[MT_ANIMALGAESEG] = 32,
	-- Castle Eggman Zone scenery
	[MT_CHAIN] = 6,
	[MT_FLAMEHOLDER] = 6,
	[MT_FIRETORCH] = 6,
	-- Arid Canyon Zone scenery
	[MT_OILLAMP] = 6,
	-- Red Volcano zone scenery
	[MT_BIGFERNLEAF] = 6,
	[MT_BIGFERN] = 16,
	[MT_JUNGLEPALM] = 6,
	[MT_TORCHFLOWER] = 6,
	[MT_WALLVINE_LONG] = 6,
	[MT_WALLVINE_SHORT] = 6,
	-- Haunted Heights Zone
	[MT_JACKO1] = 20,
	[MT_JACKO2] = 20,
	[MT_JACKO3] = 20,
	[MT_HHZSHROOM] = 6,
}



local function ResetSize(pmo)
	if not pmo then return end
	local skin = skins[pmo.skin]
	pmo.radius = skin.radius
	pmo.height = skin.radius*2
	pmo.player.camerascale = skins[pmo.skin].camerascale
	pmo.player.jumpfactor = skins[pmo.skin].jumpfactor
end

local function UpdateSize(pmo, radius)
	pmo.destradius = $ + radius/scalefactor
end


local function DropItem(k, mo)
	local o = k.object

	if o.type == MT_TOKEN then
		token = $ - 1
	end

	if o.type == MT_RING or o.type == MT_FLINGRING then
		o = P_SpawnMobj(k.object.x, k.object.y, k.object.z, MT_FLINGRING)
		o.fuse = 8*TICRATE
		k.object.type = MT_NULL
		P_KillMobj(k.object)
	else
		o.flags = o.oldflags & !MF_NOGRAVITY
		o.clumpedto = nil
		o.clumppoint = nil
	end

	local x,y,z = o.x-mo.x,o.y-mo.y,o.z-mo.z

	P_MoveOrigin(o, o.x+x, o.y+y, o.z)

	o.momx = x/5
	o.momy = y/5
	o.momz = abs(z/5)
end



local function Clump(object, pmo, size)

	local height = size or object.height
	local radius = size or object.radius

	if object.clumpedto then
		return true
	elseif pmo.state == S_PLAY_PAIN and pmo.player.powers[pw_flashing] then
		return true
	elseif pmo.player.playerstate == PST_DEAD then
		return false
	elseif IGNORE[object.type] then
		return false
	elseif not object.valid then
		return true
	elseif object.type == MT_FLINGRING and object.fuse > TICRATE*7 then
		return true
	elseif (radius + height)/2 > pmo.radius and not ALWAYS[object.type] then
		return false
	end

	object.oldflags = object.flags
	object.flags = $ | MF_NOCLIPTHING | MF_SLIDEME | MF_NOGRAVITY
	object.clumpedto = pmo

	object.clumppoint = QuatRotate({w=0,x=pmo.x-object.x,y=pmo.y-object.y,z=pmo.z-object.z+pmo.radius}, pmo.rotation)

	if object.type == MT_FLINGRING then
		object.type = MT_RING
	elseif object.type == MT_TOKEN then
		S_StartSound(mo, sfx_token)
		token = $ + 1
	end

	pmo.player.rings = $1+1

	if not pmo.katamari then pmo.katamari = {} end
	table.insert(pmo.katamari, {
		object = object,
		radius = radius,
		type = object.type
	})
	UpdateSize(pmo, radius)

	if INTERACTABLE[object.type] then
		return false
	end

	return true

end


local function AddRings(mo, rings)
	for i=1, rings do
		local x=FixedMul(cos(FixedAngle(360*FRACUNIT/rings*i)), mo.radius*2)
		local y=FixedMul(sin(FixedAngle(360*FRACUNIT/rings*i)), mo.radius*2)
		local ring = P_SpawnMobj(mo.x+x, mo.y+y, mo.z+mo.radius, MT_RING)
		Clump(ring, mo)
	end
end



local function SceneryCollision(mo)

	if mo.clumpedto then return end

	for player in players.iterate do
		if player and player.mo then
			local pmo = player.mo

			local radius = SCENERY[mo.type]*FRACUNIT
			local x = abs(pmo.x/FRACUNIT - mo.x/FRACUNIT)
			local y = abs(pmo.y/FRACUNIT - mo.y/FRACUNIT)
			local z = abs(pmo.z/FRACUNIT - mo.z/FRACUNIT)
			local dist = radius/FRACUNIT + pmo.radius/FRACUNIT
			if abs(x) < dist and abs(y) < dist and abs(z) < dist then
				Clump(mo, pmo, radius)
			end
		end
	end
end


for thing, _ in pairs(SCENERY) do
	addHook("MobjThinker", SceneryCollision, thing)
end


addHook("TouchSpecial", function(object, pmo)
	if maptol & TOL_NIGHTS > 1 then
		return
	end

	return Clump(object, pmo)
end)

addHook("MobjMoveBlocked", function(pmo, thing, line)
	if maptol & TOL_NIGHTS > 1 then
		return
	end

	if thing and thing.valid then
		return Clump(thing, pmo)
	end
end, MT_PLAYER)


addHook("BossThinker", function(boss)
	if maptol & TOL_NIGHTS > 1 then
		return
	end

	if boss.clumpedto then
		if boss.health > 0 then
			A_BossDeath(boss)
		end
	end
end)






addHook("MobjDamage", function(mo, inflictor, source, damage, damagetype)
	if maptol & TOL_NIGHTS > 1 then
		return
	end

	-- if you don't have a katamari then just behave as normal
	if not mo.katamari or #mo.katamari == 0 then
		return false
	end

	-- let instant kill damage types work as intended
	for _, d in ipairs({DMG_INSTAKILL, DMG_DROWNED, DMG_SPACEDROWN, DMG_DEATHPIT, DMG_CRUSHED})
		if damagetype == d then
			return false
		end
	end

	-- if player has a shield then let it take the hit
	if mo.player.powers[pw_shield] then
		return false
	end

	-- you can't outgrow fire or electric damage
	if damagetype != DMG_FIRE and damagetype != DMG_ELECTRIC and inflictor and Clump(inflictor, mo) then
		return true
	end

	-- if we're in knockback and flashing still don't do anything
	if mo.state == S_PLAY_PAIN or mo.player.powers[pw_flashing] > 0 then
		return true
	end

	-- drop everything
	local droppedstuff = false
	for _, k in ipairs(mo.katamari) do
		if k and k.object and k.object.valid
			droppedstuff = true
			DropItem(k, mo)
		end
	end

	-- if we had stuff to drop...
	if droppedstuff then
		mo.player.rings = 0
		P_DoPlayerPain(mo.player, source, inflictor)

		if damagetype == DMG_SPIKE then
			S_StartSound(mo, sfx_spkdth)
		else
			P_PlayRinglossSound(mo, mo.player)
		end

		return true
	end

	return false
end, MT_PLAYER)



addHook("PlayerThink", function(p)
	if maptol & TOL_NIGHTS > 1 then
		return
	end

	if p.bot > 0 then
		return
	end

	if not p.mo then return end

	local mo = p.mo

	if not mo.katamari then
		mo.katamari = {}
	end

	if not mo.destradius then
		mo.destradius = mo.radius
	end

	local dropspeed = 5
	if p.rings > 250 then
		dropspeed = 1
	elseif p.rings > 100 then
		dropspeed = 2
	elseif p.rings > 50 then
		dropspeed = 3
	elseif p.rings > 20 then
		dropspeed = 4
	end
	
	if p.dashspeed > skins[mo.skin].maxdash/2 and #mo.katamari > 0 and leveltime%5==0 then
		DropItem(mo.katamari[#mo.katamari], mo)
		p.rings = $1 - 1
	end

	local j=1
	for i=1, #mo.katamari do
		local k = mo.katamari[i]
		if k and k.object and k.object.valid and k.object.clumpedto == mo then
			local o = k.object
			local pnt = {
				w = 0,
				x = o.clumppoint.x,
				y = o.clumppoint.y,
				z = o.clumppoint.z
			}
			local tpnt = QuatRotate(pnt, mo.rotation)
			P_MoveOrigin(o, mo.x-tpnt.x, mo.y-tpnt.y, mo.z-tpnt.z+mo.radius)

			if i ~= j then
				mo.katamari[j] = mo.katamari[i]
				mo.katamari[i] = nil
			end
			j = $+1
		else
			if k and k.radius then
				UpdateSize(mo, -mo.katamari[i].radius)
			end
			mo.katamari[i] = nil
		end
	end

	if p.rings > #mo.katamari then
		AddRings(mo, p.rings - #mo.katamari)
	end
	
	p.rings = #mo.katamari

	if mo.destradius != mo.radius then

		local scaletick = FRACUNIT/12
		local diff = mo.destradius - mo.radius

		if diff > scaletick then
			if diff > 0 then
				diff = scaletick
			else
				diff = -scaletick
			end
		end

		mo.radius = $ + diff
		mo.height = $ + 2*diff
	
		local scale = FixedDiv(mo.radius, skins[mo.skin].radius)
		mo.player.shieldscale = scale

		scale = FixedSqrt(scale)
		if scale < FRACUNIT then
			scale = FRACUNIT
		end
		mo.player.camerascale = FixedMul(skins[mo.skin].camerascale, scale)
		mo.player.jumpfactor = FixedMul(skins[mo.skin].jumpfactor, scale)
	
	
		local skin = skins[mo.skin]
		if mo.radius < skin.radius then
			ResetSize(mo)
		end
	end

end)


addHook("MapLoad", function(mapnumber)
	for player in players.iterate do
		ResetSize(player.mo)
	end
end)
